<?php

namespace Drupal\lb_plus\Controller;

use Drupal\block_content\Entity\BlockContentType;
use Drupal\Component\Render\FormattableMarkup;
use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\AppendCommand;
use Drupal\Core\Ajax\MessageCommand;
use Drupal\Core\Ajax\OpenOffCanvasDialogCommand;
use Drupal\Core\Ajax\RemoveCommand;
use Drupal\Core\Block\BlockManagerInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\Form\FormBuilderInterface;
use Drupal\Core\Layout\LayoutPluginManagerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\Url;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\FieldConfigInterface;
use Drupal\file\Entity\File;
use Drupal\layout_builder\Controller\ChooseBlockController;
use Drupal\layout_builder\Controller\LayoutRebuildTrait;
use Drupal\layout_builder\Form\AddBlockForm;
use Drupal\layout_builder\LayoutTempstoreRepositoryInterface;
use Drupal\layout_builder\SectionComponent;
use Drupal\layout_builder\SectionStorageInterface;
use Drupal\lb_plus\Entity\LbPlusBlockContentSettings;
use Drupal\lb_plus\Mixin\LbPlusTrait;
use Drupal\views\Ajax\ScrollTopCommand;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Drupal\block_content\Entity\BlockContent;

class LbPlusBlockController extends ChooseBlockController  implements ContainerInjectionInterface {

  use LayoutRebuildTrait;
  use LbPlusTrait;

  /**
   * The layout tempstore repository.
   *
   * @var \Drupal\layout_builder\LayoutTempstoreRepositoryInterface
   */
  protected $layoutTempstoreRepository;

  /**
   * The UUID generator.
   *
   * @var \Drupal\Component\Uuid\UuidInterface
   */
  protected $uuidGenerator;

  /**
   * The module extension list.
   *
   * @var \Drupal\Core\Extension\ModuleExtensionList|\PHPUnit\Framework\MockObject\MockObject
   */
  protected $moduleList;

  /**
   * The form builder.
   *
   * @var \Drupal\Core\Form\FormBuilderInterface
   */
  protected $formBuilder;

  /**
   * The layout manager.
   *
   * @var \Drupal\Core\Layout\LayoutPluginManagerInterface
   */
  protected $layoutManager;

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The configuration.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * LayoutController constructor.
   *
   * @param \Drupal\layout_builder\LayoutTempstoreRepositoryInterface $layout_tempstore_repository
   *   The layout tempstore repository.
   */
  public function __construct(
    LayoutTempstoreRepositoryInterface $layout_tempstore_repository,
    UuidInterface $uuid, BlockManagerInterface $block_manager,
    EntityTypeManagerInterface $entity_type_manager,
    AccountInterface $current_user,
    ModuleExtensionList $moduleList,
    LayoutPluginManagerInterface $layout_manager,
    FormBuilderInterface $form_builder,
    ConfigFactoryInterface $config_factory
  ) {
    $this->layoutTempstoreRepository = $layout_tempstore_repository;
    $this->uuidGenerator = $uuid;
    $this->blockManager = $block_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->currentUser = $current_user;
    $this->moduleList = $moduleList;
    $this->layoutManager = $layout_manager;
    $this->formBuilder = $form_builder;
    $this->config = $config_factory->get('lb_plus.settings');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('layout_builder.tempstore_repository'),
      $container->get('uuid'),
      $container->get('plugin.manager.block'),
      $container->get('entity_type.manager'),
      $container->get('current_user'),
      $container->get('extension.list.module'),
      $container->get('plugin.manager.core.layout'),
      $container->get('form_builder'),
      $container->get('config.factory')
    );
  }

  /**
   * Returns a list of block draggable and copyable.
   *
   * @param \Drupal\layout_builder\SectionStorageInterface $section_storage
   *    The section storage
   *
   * @return array
   */
  public function chooseBlock(SectionStorageInterface $section_storage) {
    // Build the bloks list with a non-existent region.
    $build = parent::build($section_storage, 0, 'none');
    $block_categories = [
      '#type' => 'container',
      '#attributes' => ['class' => ['js-layout-builder-categories']],
    ];

    $sections = $section_storage->getSections();
    $types = $this->getAvailableBlockTypes();

    if (count($sections) < 1 || $sections[0]->getLayoutId() === 'layout_builder_blank') {
      return ['#markup' => $this->t('No blocks enabled for LB+')];
    }

    // Collects available types for layouts.
    $block_type_permissions = $this->config->get('block_type_permissions');
    $available_types = [];
    foreach ($sections as $section) {
      if (!empty($block_type_permissions[$section->getLayoutId()])) {
        foreach ($block_type_permissions[$section->getLayoutId()] as $type => $value) {
          if ($value) {
            $available_types[$type] = $type;
          }
        }
      }
    }

    foreach ($types as $type) {
      // Skip unavailable types.
      if (empty($available_types[$type->id()])) {
        continue;
      }

      $config = LbPlusBlockContentSettings::load($type->id());
      $category = $config->get('category') ?: $this->t('Default')->render();

      if (empty($block_categories[$category])) {
        $block_categories[$category] = [
          '#type' => 'details',
          '#open' => TRUE,
          '#title' => $category,
          '#attributes' => ['class' => ['js-layout-builder-category']],
        ];
      }

      $icon_path = $this->moduleList->getPath('lb_plus') . '/images/default.svg';

      if ($config->get('image') && ($file = File::load($config->get('image')))) {
          $icon_path = $file->getFileUri();
      }

      $block_categories[$category][] = [
        '#type' => 'container',
        '#attributes' => [
          'draggable' => 'true',
          'data-block_id' => $type->id(),
          'data-storage_id' => $section_storage->getStorageId(),
          'class' => [
            'lb-plus__draggable',
            'lb-plus__row',
            'js-layout-builder-block-link',
          ],
        ],
        [
          '#theme' => 'image',
          '#uri' => $icon_path,
          '#alt' => $type->label(),
          '#title' => $type->label(),
          '#attributes' => [
            'width' => '20',
            'height' => '20',
            'style' => 'width: 20px; height: 20px; overflow: hidden;',
            'class' => [
              'icon',
            ],
          ],
        ],
        [
          '#type' => 'item',
          '#markup' => $type->label(),
        ],
      ];
    }

    if (Element::children($block_categories)) {
      $build['block_types'] = $block_categories;
    }

    $build['block_types']['#weight'] = 100;
    $build['block_categories']['#weight'] = 200;

    foreach (Element::children($build['block_categories']) as $category) {
      $build['block_categories'][$category]['#open'] = FALSE;
      foreach ($build['block_categories'][$category]['links']['#links'] as &$link) {
        $link['attributes']['class'] = ['lb_plus-block_category_link', 'lb-plus__draggable', 'js-layout-builder-block-link'];
        $link['attributes']['draggable'] = 'true';
      }
    }

    // Hide the block creation.
    unset($build['add_block']);

    return $build;
  }

  /**
   * Ajax callback: add a copy icon to block.
   *
   * @param \Drupal\layout_builder\SectionStorageInterface|NULL $section_storage
   * @param $delta
   * @param $region
   * @param $uuid
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   */
  public function copyBlock(SectionStorageInterface $section_storage = NULL, $delta = NULL, $region = NULL, $uuid = NULL) {
    $response = new AjaxResponse();

    try {
      $component = $section_storage
        ->getSection($delta)
        ->getComponent($uuid);

      $plugin_id = explode(':', $component
        ->getPluginId())[1];

      $blocks = \Drupal::entityTypeManager()
        ->getStorage('block_content')
        ->loadByProperties(['uuid' => $plugin_id]);

      $block = reset($blocks);

      $delta = 0;
      if ($block) {
        $response->addCommand(new RemoveCommand('.lb_plus__paste'));
        foreach ($section_storage->getSections() as $section) {
          $layout = $section->getLayout($this->getPopulatedContexts($section_storage));
          $layout_definition = $layout->getPluginDefinition();
          foreach ($layout_definition->getRegions() as $region => $info) {
            $selector = "block-$delta-$region";
            $element = [
              '#type' => 'link',
              '#title' => t('Paste'),
              '#url' => Url::fromRoute('lb_plus.paste_block', [
                'section_storage_type' => $section_storage->getStorageType(),
                'section_storage' => $section_storage->getStorageId(),
                'region' => $region,
                'delta' => $delta,
                'plugin_id' => $block->id(),
              ]),
              '#attributes' => [
                'class' => ['use-ajax', 'lb_plus__paste'],
              ],
            ];

            $response->addCommand(new AppendCommand('div[data-layout-builder-highlight-id^="' . $selector . '"]', \Drupal::service('renderer')->render($element)));
          }
          $delta++;
        }
      }

    } catch (\Exception $e) {}

    return $response;
  }

  /**
   * Ajax callback: create a new block on drop.
   *
   * @param \Drupal\layout_builder\SectionStorageInterface|NULL $section_storage
   * @param $delta
   * @param $region
   * @param $plugin_id
   *
   * @return mixed
   */
  public function addBlock(SectionStorageInterface $section_storage = NULL, $delta = NULL, $region = NULL, $plugin_id = NULL) {
    // Check if current block type can be dropped into current layout.
    $block_type_permissions = $this->config->get('block_type_permissions');
    $layout_id = $section_storage->getSection($delta)->getLayoutId();
    $access = !empty($block_type_permissions[$layout_id][$plugin_id]);

    if ($access) {
      $block = BlockContent::create([
        'info' => new FormattableMarkup('Block @plugin_id for @region region of @label', [
          '@plugin_id' => $plugin_id,
          '@region' => $region,
          '@label' => $section_storage->label(),
        ]),
        'type' => $plugin_id,
      ]);

      // Generate sample data for required fields without default value.
      foreach ($block->getFields(FALSE) as $field) {
        $data_definition = $field->getDataDefinition();
        $field_definition = $field->getFieldDefinition();
        if ($data_definition instanceof FieldConfig) {
          $default_value = NULL;
          // Image field special case.
          if ($field_definition instanceof FieldConfigInterface && $field_definition->getType() == 'image') {
            $settings = $field_definition->getSetting('default_image');
            $default_value = !empty($settings['uuid']);
          }
          else {
            $default_value = $field->getValue();
          }
          if ($data_definition->get('required') && empty($default_value)) {
            $field->generateSampleItems();
          }
        }
      }

      $block->setReusable();

      try {
        if ($block->save()) {
          $section = $section_storage->getSection($delta);
          $component = new SectionComponent($this->uuidGenerator->generate(), $region, ['id' => 'block_content:' . $block->uuid()]);
          $section->appendComponent($component);
          $this->layoutTempstoreRepository->set($section_storage);
        }
      } catch (\Exception $e) {
      }

      return $this->rebuildLayout($section_storage);
    }
    else {
      $response = new AjaxResponse();
      $block_type = BlockContentType::load($plugin_id);
      $layout = $section_storage->getSection($delta)->getLayout();
      $message = $this->t('The "@block_type" block type is not permitted for the "@section_type" layout section.', [
        '@block_type' => $block_type->label(),
        '@section_type' => $layout->getPluginDefinition()->getLabel(),
      ]);
      $selector = '#node-page-layout-builder-form';
      $response->addCommand(new MessageCommand($message, $selector, ['type' => 'error'], TRUE));
      $response->addCommand(new ScrollTopCommand($selector));
      return $response;
    }
  }


  /**
   * Ajax callback: paste a copied block.
   *
   * @param \Drupal\layout_builder\SectionStorageInterface|NULL $section_storage
   * @param $delta
   * @param $region
   * @param $plugin_id
   *
   * @return void
   */
  public function pasteBlock(SectionStorageInterface $section_storage = NULL, $region = NULL, $delta = NULL, $plugin_id = NULL) {
    $block = BlockContent::load($plugin_id);

    if ($block && $region && $delta) {
      $cloned_entity = $block->createDuplicate();
      $label_key = \Drupal::entityTypeManager()
        ->getDefinition('block_content')
        ->getKey('label');

      $new_label = str_replace(LB_PLUS_TEMPLATE_PLACEHOLDER, '', $block->label());

      $cloned_entity->set($label_key, trim($new_label));
      $cloned_entity->setReusable();

      if ($cloned_entity->save()) {
        $section = $section_storage->getSection($delta);
        $component = new SectionComponent($this->uuidGenerator->generate(), $region, ['id' => 'block_content:' . $cloned_entity->uuid()]);
        $section->appendComponent($component);
        $this->layoutTempstoreRepository->set($section_storage);
      }
    }

    return $this->rebuildLayout($section_storage);
  }

  /**
   * Ajax callback: add a default block.
   *
   * @param \Drupal\layout_builder\SectionStorageInterface|NULL $section_storage
   * @param $region
   * @param $delta
   * @param $plugin_id
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   */
  public function addDefaultBlock(SectionStorageInterface $section_storage = NULL, $region = NULL, $delta = NULL, $plugin_id = NULL) {
    $form = $this->formBuilder->getForm(AddBlockForm::class, $section_storage, $delta, $region, $plugin_id);
    $response = new AjaxResponse();
    $response->addCommand(new OpenOffCanvasDialogCommand('form', $form));
    return $response;
  }

  /**
   * Get the block type categories.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   */
  public function getCategories() {
    $configs = LbPlusBlockContentSettings::loadMultiple();

    $suggestions = [];
    $results = ['Default'];

    $query = \Drupal::request()->get('q');

    if ($query) {
      $results[$query] = $query;

      foreach ($configs as $config) {
        $category = $config->get('category');
        if (substr(strtolower($category), 0, strlen($query)) === strtolower($query)) {
          $results[$category] = $category;
        }
      }
    }

    foreach ($results as $result) {
      $suggestions[] = [
        'value' => $result,
        'label' => $result,
      ];
    }

    return new JsonResponse($suggestions);
  }
}
